/* bzflag
 * Copyright (c) 1993-2025 Tim Riker
 *
 * This package is free software;  you can redistribute it and/or
 * modify it under the terms of the license found in the file
 * named COPYING that should have accompanied this file.
 *
 * THIS PACKAGE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

// Template.cxx
//   A short description of the module's function


#include "common.h"

// interface header
#include "Template.h"

// system headers

// common headers

// local headers



/*
 * Local Variables: ***
 * mode: C++ ***
 * tab-width: 4 ***
 * c-basic-offset: 4 ***
 * indent-tabs-mode: nil ***
 * End: ***
 * ex: shiftwidth=4 tabstop=4
 */
