/* bzflag
 * Copyright (c) 1993-2025 Tim Riker
 *
 * This package is free software;  you can redistribute it and/or
 * modify it under the terms of the license found in the file
 * named COPYING that should have accompanied this file.
 *
 * THIS PACKAGE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef __CUSTOM_BOX_H__
#define __CUSTOM_BOX_H__

/* interface header */
#include "WorldFileObstacle.h"

/* common interface headers */
#include "BzMaterial.h"

/* local interface headers */
#include "WorldInfo.h"


class CustomBox : public WorldFileObstacle
{
public:
    CustomBox();
    ~CustomBox();
    bool read(const char* cmd, std::istream& input) override;
    void writeToGroupDef(GroupDefinition*) override;

private:
    enum
    {
        XP = 0,
        XN,
        YP,
        YN,
        ZP,
        ZN,
        FaceCount
    };

    bool isOldBox;

    int phydrv[FaceCount];
    float texsize[FaceCount][2];
    float texoffset[FaceCount][2];
    bool drivethrough[FaceCount];
    bool shootthrough[FaceCount];
    bool ricochets[FaceCount];
    BzMaterial materials[FaceCount];

    static const char* faceNames[FaceCount];
};


#endif  /* __CUSTOM_BOX_H__ */

// Local variables: ***
// mode: C++ ***
// tab-width: 4***
// c-basic-offset: 4 ***
// indent-tabs-mode: nil ***
// End: ***
// ex: shiftwidth=4 tabstop=4
