
//
// wavefrontOBJ.h : definitions of the model classes
//

#ifndef _WAVEFRONT_OBJ_H_
#define _WAVEFRONT_OBJ_H_

#include "model.h"
#include <string>

void readOBJ ( CModel &model, std::string file );

#endif //_WAVEFRONT_OBJ_H_

