..\..\man\man2html.exe < ..\..\man\bzw.5.in > ..\..\man\bzw.html
..\..\man\man2html.exe < ..\..\man\bzflag.6.in > ..\..\man\bzflag.html
..\..\man\man2html.exe < ..\..\man\bzadmin.6.in > ..\..\man\bzadmin.html
..\..\man\man2html.exe < ..\..\man\bzfs.6.in > ..\..\man\bzfs.html
