/* bzflag
 * Copyright (c) 1993-2025 Tim Riker
 *
 * This package is free software;  you can redistribute it and/or
 * modify it under the terms of the license found in the file
 * named COPYING that should have accompanied this file.
 *
 * THIS PACKAGE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef __WEAPON_H__
#define __WEAPON_H__

// BZFlag common header
#include "common.h"

#include <vector>
#include "Flag.h"

class Weapon
{
public:
    Weapon();
    const void* unpack(const void*);

public:
    FlagType *type;
    float pos[3];
    float dir;
    float initDelay;
    std::vector<float> delay;
};


#endif /* __WEAPON_H__ */

// Local Variables: ***
// mode: C++ ***
// tab-width: 4 ***
// c-basic-offset: 4 ***
// indent-tabs-mode: nil ***
// End: ***
// ex: shiftwidth=4 tabstop=4
